/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <gdbm.h>

#include <mplib1/dl_list.h>
#include <mplib1/fprintfile.h>
#include <mplib1/cfg_file.h>

#include <mplib1/gdbm_util.h>


/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

static char util_debug_str[]="GDBM_UTIL_DEBUG";
static int gdbm_util_debug=0;

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static void
gdbm_fail( const char *errstr )
{
    fprintfile( stderr,"GDBM Error: %s\n", errstr );
    return;
}

int
gdbm_db_close( struct gdbm_help *gptr )
{
    gdbm_util_debug=get_config_flag(util_debug_str);
    if (gdbm_util_debug)
	fprintfile(stderr,"gdbm_db_close: Close (%s)\n", gptr->name );
    if ( *gptr->file_ptr )
    {
	gdbm_close( *gptr->file_ptr );
	if (gdbm_util_debug)
	    fprintfile(stderr,"gdbm_db_close: Closed\n" );
	*gptr->file_ptr = NULL;
    }
    else if (gdbm_util_debug)
	fprintfile(stderr,"gdbm_db_close: not open\n" );
    return(1);
}

static GDBM_FILE
gdbm_open_this( const char *name, int blk_sz, int o_flags,
		int cache_sz, void (*fatal_func)(const char *) )
{
    GDBM_FILE rv=NULL;
 
    if (gdbm_util_debug)
	fprintfile(stderr,"gdbm_open_this: %s %d %x\n", name, blk_sz, o_flags );
    rv = gdbm_open( (char *)(size_t)name, blk_sz, o_flags, 0600, fatal_func );
    if (rv && cache_sz)
    {
	gdbm_setopt( rv, GDBM_CACHESIZE, &cache_sz, sizeof(int) );
    }
    return(rv);
}

static int
gdbm_open_one( const char *cfg_name, const char *def_name,
		const char *block_size, const char *non_buf,
		const char *cache_size, int force_new,
		void (*fatal_func)(const char *), GDBM_FILE *p_rv )
{
    GDBM_FILE rv=NULL;
    int o_flags;
    int block_sz=0,cache_sz=0;
    char *db_name;
 
    /* Do this one */
    db_name = strdup( eval_config_default( cfg_name, def_name ));
    o_flags= ((force_new)?GDBM_NEWDB : GDBM_WRCREAT);
    if (non_buf && get_config_flag(non_buf))
	o_flags |= GDBM_FAST;
    if (block_size)
	block_sz = get_config_int( block_size );
    if (cache_size)
	cache_sz = get_config_int( cache_size );
    if (fatal_func==NULL)
	fatal_func = gdbm_fail;
    rv = gdbm_open_this( db_name, block_sz, o_flags, cache_sz, fatal_func );
    if (rv==NULL)
    {
	fprintfile(stderr,"Unable to open database file (%s)\n", db_name );
    }
    free(db_name);
    *p_rv = rv;
    return((rv==NULL)?0:1);
}

int
gdbm_db_open( struct gdbm_help *gptr, int force_new )
{
    int rv=0;

    gdbm_util_debug=get_config_flag(util_debug_str);
    if (gptr->name)
    {
	if (   gptr->conditional==NULL
	    || !get_config_flag(gptr->conditional)
	   )
	{
	    rv = gdbm_open_one( gptr->name, gptr->def_name,
				gptr->blocksize, gptr->fast,
				gptr->cachesize, force_new,
				gptr->fatal_func, gptr->file_ptr );
	    if (gdbm_util_debug)
		fprintfile(stderr,"gdbm_db_open: open - rv = %d\n", rv );
	}
	else
	{
	    rv = 1;
	    if (gdbm_util_debug)
		fprintfile(stderr,"gdbm_db_open: open - conditional\n" );
	}
    }else
    if (gdbm_util_debug)
	fprintfile(stderr,"gdbm_db_open: open - no name\n" );
    return(rv);
}

int
gdbm_db_sync_set( struct gdbm_help *gptr, int sync_mode )
{
    int sv;
    sv = (sync_mode) ? 1:0;
    if (*gptr->file_ptr)
	gdbm_setopt( *gptr->file_ptr, GDBM_FASTMODE, &sv, sizeof(sv) );
    return(1);
}

char *
gdbm_key_gen( char *buf, const char *format, ... )
{
    va_list ap;

    if (buf==NULL)
	buf=malloc(255);
    if (buf)
    {
	va_start( ap, format );
	vsprintf( buf, format, ap );
	va_end( ap );
    }
    return(buf);
}

int
store_some_record( GDBM_FILE file,
		    const void *ptr,
		    size_t psize,
		    const char *format, ... )
{
    datum skey,sdata;
    char some_key[255];
    va_list ap;
    int rv;

    gdbm_util_debug=get_config_flag(util_debug_str);
    va_start( ap, format );
    vsprintf( some_key, format, ap );
    va_end( ap );
    if (gdbm_util_debug)
	fprintfile(stderr,"store_some_record: %s\n", some_key );
    skey.dptr = some_key;
    skey.dsize = (int)strlen(some_key) + 1;
    sdata.dptr = (char *)(size_t)ptr;
    sdata.dsize = (int)psize;
    rv = gdbm_store( file, skey, sdata, GDBM_REPLACE );
    if (gdbm_util_debug)
	fprintfile(stderr,"store_some_record: rv=%p\n", rv );
    return( rv );
}

void *
fetch_some_record( GDBM_FILE file, const char *format, ... )
{
    datum skey,sdata;
    char some_key[255];
    va_list ap;

    gdbm_util_debug=get_config_flag(util_debug_str);
    va_start( ap, format );
    vsprintf( some_key, format, ap );
    va_end( ap );
    if (gdbm_util_debug)
	fprintfile(stderr,"fetch_some_record: %s\n", some_key );
    skey.dptr = some_key;
    skey.dsize = (int)strlen(some_key) + 1;
    sdata = gdbm_fetch( file, skey );
    if (gdbm_util_debug)
	fprintfile(stderr,"fetch_some_record: rv=%p\n", sdata.dptr );
/*    DO_AREP( sdata.dptr, sdata.dsize );*/
    return( (void *)(sdata.dptr) );
}

int
delete_some_record( GDBM_FILE file, const char *format, ... )
{
    datum skey;
    int rv;
    char some_key[255];
    va_list ap;

    gdbm_util_debug=get_config_flag(util_debug_str);
    va_start( ap, format );
    vsprintf( some_key, format, ap );
    va_end( ap );
    if (gdbm_util_debug)
	fprintfile(stderr,"delete_some_record: %s\n", some_key );
    skey.dptr = some_key;
    skey.dsize = (int)strlen(some_key) + 1;
    rv = gdbm_delete( file, skey );
    if (gdbm_util_debug)
	fprintfile(stderr,"delete_some_record: rv=%d\n", rv );
    return( rv );
}



/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

struct walk_det
{
FILE		*fh;
GDBM_FILE	gfile;
int		(* dfunc)(FILE *, char *, void *);
int		do_data;
int		flags;
};


/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static void
disp_item( dl_Node_t *nptr, struct walk_det *wdp )
{
    datum skey,sdata;
    int rv;

    skey.dptr = (char *)nptr->ln_Name;
    skey.dsize = (int)strlen( nptr->ln_Name ) + 1;
    if (wdp->dfunc)
    {
	if (wdp->do_data)
	{
	    sdata = gdbm_fetch( wdp->gfile, skey );
	    rv = (*wdp->dfunc)( wdp->fh, skey.dptr, sdata.dptr );
	    if (sdata.dptr && (rv & GDBM_IDUMP_KEEP_DATA)==0)
		free(sdata.dptr);
	}
	else
	    (*wdp->dfunc)( wdp->fh, skey.dptr, NULL );
    }
    else
	fprintf( wdp->fh, "Key: %s\n", skey.dptr );
    return;
}

static void
free_item( dl_Node_t *nptr )
{
    /* Free the string and the node */
    free( (char *)nptr->ln_Name );
    free( nptr );
    return;
}

static int
add_to_list( dl_List_t *lp, char *key )
{
    dl_Node_t *nptr;

    nptr = malloc( NODE_SIZE );
    if (nptr)
    {
	dl_Init_Node( nptr, key, nptr );
	dl_Add_Tail( lp, nptr );
    }
    return((nptr!=NULL));
}


int
gdbm_index_dump( FILE *fp,
		struct gdbm_help *gptr,
		int flags,
		gdbm_index_dump_t dfunc )
{
    GDBM_FILE gfile;
    datum skey,nextkey;
    int do_sort, do_hdr, do_data;
    struct walk_det walk_details;
    dl_List_t sort_list;
    dl_Node_t tnode;


    gfile = *gptr->file_ptr;
    do_hdr = (flags & GDBM_IDUMP_HDR);
    do_sort = (flags & GDBM_IDUMP_SORT);
    do_data = (flags & GDBM_IDUMP_DATA);
    dl_Init_List( &sort_list, 0 );

    walk_details.fh = fp;
    walk_details.gfile = gfile;
    walk_details.dfunc = dfunc;
    walk_details.do_data = do_data;
    walk_details.flags = flags;

    if (do_hdr)
	fprintf( fp, "   Index dump of %s\n", gptr->name );

    skey = gdbm_firstkey ( gfile );
    while ( skey.dptr )
    {
	if (do_sort)
	{
	    add_to_list( &sort_list, skey.dptr );
	}else
	{
	    tnode.ln_Name = skey.dptr;
	    disp_item( &tnode, &walk_details );
	}
	nextkey = gdbm_nextkey ( gfile, skey );
	if (do_sort==0)
	    free ( skey.dptr );
	skey = nextkey;
    };
    if (do_sort)
    {
	dl_sort_n( &sort_list );
	dl_Walk_List( &sort_list, (dl_Walk_List_t)disp_item, &walk_details );
	dl_Walk_List( &sort_list, (dl_Walk_List_t)free_item, NULL );
    }
    if (do_hdr)
	fprintf( fp, "\n" );
    return(0);
}


/* -- End of File -- */
