/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#ifdef __linux__
#include <getopt.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <signal.h>

#include <mplib1/dl_list.h>
#include <mplib1/pid_check.h>
#include <mplib1/fprintfile.h>
#include <mplib1/cfg_file.h>
#include "include/bpo_internal.h"
#include <mplib1/bpo.h>

#include "sh_diag.h"

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

extern int interactive_loop( void *vp );

static char e_str[]="<NULL>";
static int clean_up=0;
static dl_List_t plist;

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static void usage ( const char *progname, const char *errstr )
{
    if (errstr)
	fprintf( stderr, "Error: %s\n", errstr );
    fprintf(stderr,
	    "\nUsage: %s [ -f filename ] [ -s | -e | -d ] [ -v ]\n", progname );
    exit(1);
    return;
}

static void
dump_list( dl_Node_t *np )
{
    fputs( np->ln_Name, stdout );
    dl_Free_Node( np, DL_FREE_NAMES | DL_FREE_NODES );
    return;
}

static int
mpprintf( const char *format, ... )
{
    va_list ap;
    char bigbuf[4096];
    char *cp;
    dl_Node_t *tn;

    va_start( ap, format );
    vsprintf( bigbuf, format, ap );
    va_end(ap);

    cp=strdup(bigbuf);
    if (cp)
    {
	tn = malloc( NODE_SIZE );
	if (tn)
	{
	    dl_Init_Node( tn, cp, tn );
	    dl_Add_Tail( &plist, tn );
	}else
	    free(cp);
    }
    return(0);
}

static int
print_list_details( void *seg_base,
		    struct shalloc_track_head *sthp,
		    const char *lname )
{
    struct shalloc_track *stp;
    int t;

    mpprintf( "List: %s\n", lname );
    while (sthp)
    {
	stp = GPTR( sthp, sizeof(struct shalloc_track_head));
	mpprintf( "sth @ %lx %d entries, %d empty\n",
		sthp->lno_me, sthp->num_tracks, sthp->num_empty );
	for (t=0; t< sthp->num_tracks; t++)
	{
	    if (stp->sh_off != 0)
		mpprintf( "  block @ %lx size %x\n",
			stp->sh_off, stp->sh_size );
	    stp++;
	};
	sthp = GRPTR( sthp, sthp->next_track_head );
    }
    return(0);
}

static int
do_checks( void *where, int which )
{
    /* First we check that the hint is in the permitted region */
    char *seg_base;
    struct shalloc_base *sh_base;
    struct shalloc_track_head *free_head,*used_head;

    seg_base = Find_SODB_Resource( where, SODB_BASE_NAME );
    if (seg_base == NULL)
	return(0);

    sh_base = Find_SODB_Resource( seg_base, "shalloc" );
    mpprintf( " show shalloc_base %08lx\n", sh_base->lno_me );
    if (sh_base==NULL)
	return(0);

    /* well, let's lock this thing for our own use */
    b_Lock( &sh_base->sb_lock );

    /* OK, it's all ours. What shall we do ? */
    free_head = GRPTR( sh_base, sh_base->lno_first_free );
    used_head = GRPTR( sh_base, sh_base->lno_first_used );

    if (which & BIT(0))
	print_list_details( seg_base, free_head, "Free List" );
    if (which & BIT(1))
	print_list_details( seg_base, used_head, "Used List" );

    b_Unlock( &sh_base->sb_lock );
    return(1);
}

int
diag_Validate_Shalloc( void *where, int which, int do_clean )
{
    clean_up = do_clean;
    do_checks( where, which );
    /* Now dump list contents */
    dl_Walk_List( &plist, (dl_Walk_List_t)dump_list, NULL );
    return(0);
}

static const char *
nstr( const char *cp )
{
    return( (char *)(size_t)(cp==NULL)?e_str:cp );
}

static int
proc_list( struct shm_process_private *sppp, void *param1 )
{
    int ks;
    ks = is_pid_dead(sppp->pid);
    mpprintf("%08lx %-12.12s %s  <%s> <%s> <%s>\n",
	    sppp->spp->proc_Node.lno_me,
	    sppp->pid_str,
	    (const char *)((ks==0) ? "Alive" : "Dead "),
	    nstr( sppp->pid_grp ),
	    nstr( sppp->pid_nm ),
	    nstr( sppp->pid_fifo ) );
    if (clean_up && ks)
	return(SHM_DELETE_PROC);
    return(SHM_LEAVE_PROC);
}

int
diag_Validate_Processes( void *vp, int do_clean )
{
    clean_up = do_clean;
    mpprintf("Processes - show shm_process\n");
    mpprintf("Offset   pid          State Group/Program/Fifo\n" );
    Validate_Processes( vp, SHM_CHECK_ALL_PROCESSES, proc_list, NULL );
    /* Now dump list contents */
    dl_Walk_List( &plist, (dl_Walk_List_t)dump_list, NULL );
    return(0);
}


static void
q_pid( struct bpo_private_pid *ppp, int *any_alive )
{
    int q_dead;

    q_dead = is_pid_dead(ppp->pid);
    mpprintf("                                %-8ld  %s\n",
	    (long)ppp->pid,
	    (q_dead)?"Dead":""  );
    if (q_dead==0)
	(*any_alive)++;
    return;
}


static int
q_list( struct bpo_private_q *pqp, void *param1 )
{
    int rv = Q_LEAVE_QUEUE;

    if (pqp->Q_notify == Q_NM_PIPE)
    {
	if (pqp->Q_flags & Q_SOLE_READER)
	{
	    char tstr[12];
	    gen_pid_str( tstr, pqp->Q_sole_pid );
	    mpprintf("%08lx  %-4d  %-4d  %08lx  %-8ld  %s\n",
		    pqp->qp->Q_Node.lno_me, pqp->qp->Q_List.lno_count,
		    pqp->Q_max_count, 0,
		    (long)pqp->Q_sole_pid, pqp->Q_name );
	    if (   clean_up
		&& (pqp->Q_flags & Q_PERSISTENT)==0
		&& is_pid_dead(pqp->Q_sole_pid)
	       )
		rv = Q_DELETE_QUEUE;
	}else
	{
	    int any_alive;
	    mpprintf("%08lx  %-4d  %-4d  %08lx  PipeM     %s\n",
		    pqp->qp->Q_Node.lno_me, pqp->qp->Q_List.lno_count,
		    pqp->Q_max_count, 0, pqp->Q_name );
	    any_alive=0;
	    dl_Walk_List( &pqp->private_pids,
			(dl_Walk_List_t)q_pid,
			&any_alive );
	    if (   any_alive==0
		&& (pqp->Q_flags & Q_PERSISTENT)==0
		&& clean_up)
		rv = Q_DELETE_QUEUE;
	}
    }else
    {
	mpprintf("%08lx  %-4d  %-4d  %08lx            %s\n",
		pqp->qp->Q_Node.lno_me, pqp->qp->Q_List.lno_count,
		pqp->Q_max_count, 0, pqp->Q_name );
    }
    return(rv);
}

int
diag_Validate_Queues( void *vp, int do_clean )
{
    clean_up = do_clean;
    mpprintf("Queues - show bpo_q\nAddress   Count Max   Lock      Notify    Name\n");
    Validate_Queues( vp, Q_CHECK_ALL_QUEUES, q_list, NULL );
    /* Now dump list contents */
    dl_Walk_List( &plist, (dl_Walk_List_t)dump_list, NULL );
    return(0);
}

static int
res_list( const char *res_name, off_t res_off, void *param1 )
{
/*    mpprintf("Resource: %s %08x\n", res_name, res_off );*/
    mpprintf("%08lx %s\n", res_off, res_name );
    return(0);
}

int
diag_Validate_Resources( void *vp, int do_clean )
{
    clean_up = do_clean;
    mpprintf("Resources:\n");
    Validate_Resources( vp, CHECK_ALL_RESOURCES,
			res_list,
			NULL );
    /* Now dump list contents */
    dl_Walk_List( &plist, (dl_Walk_List_t)dump_list, NULL );
    return(0);
}

static void *
Attach_Raw_IPC( int shmid )
{
    void *vp;

    vp = shmat( shmid, NULL, 0 );

    if (vp== (void *)-1)
	vp =NULL;
    return(vp);
}


int
main (int argc, char *argv[] )
{
    int rv=0;
    int optc;
    char *fname=NULL;
    char *seg_name=NULL;
    size_t sz=(size_t)0;
    char id=0;
    void *vp;
    int do_shm=0;
    int do_mmap=0;
    int do_image=0;
    int do_rawipc=0,shmid;
    int shalloc=0,process=0,queues=0,resources=0;

    dl_Init_List( &plist, 0 );

    while ((optc = getopt(argc, argv, "CSPRQd:n:f:m:i:s:g:r:")) != EOF)
    switch (optc)
    {
    case 'r' : /* raw IPC */
	shmid = atoi(optarg);
	do_rawipc=1;
	break;
    case 'g' : /* grabbed file */
	fname = optarg;
	do_image = 1;
	break;
    case 'n' : /* name of segment */
	seg_name = optarg;
	break;
    case 'm' : /* shared memory use */
	fname = optarg;
	do_shm = 1;
	break;
    case 'f' : /* shared file */
	fname = optarg;
	do_mmap = 1;
	break;
    case 'i' : /* id */
	id = (char)atoi(optarg);
	break;
    case 's' : /* size */
	sz = (size_t)atoi(optarg);
	break;
    case 'd' : /* defaults */
	read_config_file( optarg );
	break;
    case 'C' : clean_up=1;break;
    case 'S' : shalloc=1; break;
    case 'P' : process=1;break;
    case 'Q' : queues=1;break;
    case 'R' : resources=1;break;
    default:
	usage( argv[0], "Invalid argument supplied" );
	break;
    }

    if (do_rawipc)
    {
	printf( "Attach_Raw_IPC( %d )\n", shmid );
	vp = Attach_Raw_IPC( shmid );
    }
    else if (do_image)
    {
	printf( "Attach_SODB_image( %s )\n", fname );
	vp = Attach_SODB_image( fname );
    }
    else if (do_shm)
    {
	printf( "Create_SODB_shm( %s, %s, %d %zd, 0, SODB_USE_SHALLOC )\n",
		(seg_name)?seg_name:"<as file name>", fname, id, sz );
	vp = Create_SODB_shm( seg_name, fname, id, sz, SODB_USE_SHALLOC );
    }else if (do_mmap)
    {
	printf( "Create_SODB_mmap( %s, %s, %zd, 0, SODB_USE_SHALLOC )\n",
		(seg_name)?seg_name:"<as file name>", fname, sz );
	vp = Create_SODB_mmap( seg_name, fname, sz, 0, SODB_USE_SHALLOC );
    }else
    {
	vp = comms_shm_attach( eval_config_default( "SHM_CONFIG", "$HOME/config/shm.cfg" ), SODB_READ_CONFIG );
    }

    printf("Create returned: %p\n", vp );
    if(vp)
    {
	if (shalloc || resources || process || queues)
	{
	    if (shalloc)
		do_checks( vp, 255 );
	    if (resources)
		Validate_Resources( vp, CHECK_ALL_RESOURCES, res_list, NULL );
	    if (process)
		Validate_Processes( vp, SHM_CHECK_ALL_PROCESSES, proc_list, NULL );
	    if (queues)
		Validate_Queues( vp, Q_CHECK_ALL_QUEUES, q_list, NULL );
	    /* Now dump list contents */
	    dl_Walk_List( &plist, (dl_Walk_List_t)dump_list, NULL );
	}else
	{
	    interactive_loop( vp );
	}
    }
    return(rv);
}

/* -- End of File -- */
