/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole
*******************************************************************************
**
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#ifdef __linux__
#include <getopt.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>

#include <mplib1/dl_list.h>
#include <mplib1/pid_check.h>
#include <mplib1/vre.h>
#include <mplib1/timestamp.h>
#include "include/bpo_internal.h"
#include <mplib1/bpo.h>
#include <mplib1/data_align.h>

#include "sh_diag.h"

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

static char unl_str[]="Unlocked";
static char lck_str[]="Locked";

#define lno_string(s,b) ((const char *)((s)?(Get_SODB_Pointer(b,s)):"\0"))

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static const char *
stack_prefix( struct vre_disp_walk_item *dwip, const char *suffix )
{
    /* allocate and build a new string consisting of the current
	prefix plus the suffix, set the prefix to the new
	string and return the old prefix for later restoration
    */
    char *new_prefix;
    const char *old_prefix;
    old_prefix = dwip->prefix;
    new_prefix = malloc( strlen( old_prefix ) + strlen( suffix ) + 1 );
    if (new_prefix)
    {
	strcpy( new_prefix, old_prefix );
	strcat( new_prefix, suffix );
    }
    dwip->prefix = new_prefix;
    return(old_prefix);
}

static const char *
unstack_prefix( struct vre_disp_walk_item *dwip, const char *old_prefix )
{
    /* Undo a previous allocation */
    if (dwip->prefix)
	free( (void *)(dwip->prefix) );
    dwip->prefix = old_prefix;
    return(old_prefix);
}

static char *
b_lock_str( b_lock_t *lockp )
{
    return( (b_Lockval(lockp))?lck_str:unl_str );
}

static int
b_lock_disp( b_lock_t *lockp, struct vre_disp_walk_item *dwip )
{
    fprintf( dwip->fp, "%s", b_lock_str( lockp ) );
    return(0);
}

static int
bpo_pid_lock_disp( struct bpo_pid_lock *bplp,
		    struct vre_disp_walk_item *dwip )
{
    long lstate;
    char pidstr[PID_STR_LEN];
    lstate = b_Lockval( &bplp->bpo_pid_lock );
    if (lstate)
    {
	gen_pid_str( pidstr, bplp->lock_pid ),
	fprintf( dwip->fp, "%s by %s %d times\n",
		lck_str, pidstr, bplp->pid_cnt );
	if (dwip->flags & VRE_DISP_ELEM_NAMES)
	    vre_Disp_Prefix( dwip->fp, bplp,
			    dwip->base, dwip->prefix, NULL );
	fprintf( dwip->fp,"locked at %s",
		build_timestamp_str( &bplp->lock_time, NULL ) );
    }else
    {
	fprintf( dwip->fp, unl_str );
    }
    if (bplp->unlock_pid)
    {
	gen_pid_str( pidstr, bplp->unlock_pid ),
	fprintf( dwip->fp, "\nBeing unlocked by %s", pidstr );
    }
    return(0);
}

static int
bpo_node_disp( bpo_Node_t *np, struct vre_disp_walk_item *dwip )
{
    /* Display the details of this node */
    fprintf( dwip->fp, "Me=%08lx  Fw=%08lx  Bk=%08lx\n",
	    np->lno_me, np->lno_Succ, np->lno_Pred );
    if (dwip->flags & VRE_DISP_ELEM_NAMES)
	vre_Disp_Prefix( dwip->fp, np,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Lt=%08lx  It=%08lx  Nm=%08lx(%s)",
	    np->lno_List,  np->lno_Item, np->lno_Name,
	    lno_string( np->lno_Name, np ) );
    return(0);
}

static int
newline( FILE *fp )
{
    return(fputc( '\n', fp ));
}

static int
bpo_list_disp( bpo_List_t *lp, struct vre_disp_walk_item *dwip )
{
    int pref;
    const char *cp;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    /* display details of list structure */
    fprintf( dwip->fp, "Me=%08lx\n", lp->lno_me );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &lp->lno_Tail,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Count=%d  SerNbr=%ld  Case-%sensitive\n",
	    lp->lno_count, lp->lno_serial_nbr,
	    (lp->lno_Flags& LN_IGNORECASE)?"Ins":"S" );

    cp = stack_prefix( dwip, ".lno_Head" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &lp->lno_Head,
			dwip->base, dwip->prefix, NULL );
    bpo_node_disp( &lp->lno_Head, dwip );
    newline( dwip->fp );
    cp = unstack_prefix( dwip, cp );
    cp = stack_prefix( dwip, ".lno_Tail" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &lp->lno_Tail,
			dwip->base, dwip->prefix, NULL );
    bpo_node_disp( &lp->lno_Tail, dwip );
    newline( dwip->fp );
    cp = unstack_prefix( dwip, cp );
    cp = stack_prefix( dwip, ".lno_Pid_Lock" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &lp->lno_Pid_Lock,
			dwip->base, dwip->prefix, NULL );
    bpo_pid_lock_disp( &lp->lno_Pid_Lock, dwip );
    cp = unstack_prefix( dwip, cp );
    return(0);
}

static int
sodb_resource_disp( struct sodb_resource *resp,
		    struct vre_disp_walk_item *dwip )
{
    int pref;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    bpo_node_disp( &resp->res_node, dwip );
    newline( dwip->fp );
    if (pref)
	vre_Disp_Prefix( dwip->fp, resp,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Resource at %08lx", resp->res_off );
    return(0);
}

static char *
sane_char( char tc )
{
    static char tbuf[10];
    /* munge the character into something sane for printing */
    if(tc<' ' || tc>0x7e)
    {
	sprintf(tbuf,"0x%02x",(int)(unsigned char)tc);
    }else
    {
	tbuf[0]='\'';
	tbuf[1]=tc;
	tbuf[2]='\'';
	tbuf[3]='\0';
    }
    return(tbuf);
}

static int
sodb_disp( struct sodb *sodbp,
	    struct vre_disp_walk_item *dwip )
{
    static char mmap_str[]=".mmap";
    static char ipc_str[]=".ipc";
    int pref;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Type=%d  Version=%d  Size=%zd\n",
	    sodbp->sodb_type, sodbp->sodb_version, sodbp->sodb_size );
    if (pref)
	vre_Disp_Prefix( dwip->fp, sodbp,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, (sodbp->sodb_flags & SODB_USE_SHALLOC)?
	    "Flags=SODB_USE_SHALLOC\n":"Flags=\n" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, sodbp,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Use Name=%s\n", sodbp->sodb_use_name );
    if (sodbp->sodb_type & SODB_T_MMAP)
    {
	if (pref)
	    vre_Disp_Prefix( dwip->fp, sodbp,
			    dwip->base, dwip->prefix, mmap_str );
	fprintf( dwip->fp, "fd=%d  mode=%d  flags=%d  offset=%d\n",
		sodbp->mmap_fd, sodbp->mmap_prot,
		sodbp->mmap_flags, (int)sodbp->mmap_off );
	if (pref)
	    vre_Disp_Prefix( dwip->fp, sodbp,
			    dwip->base, dwip->prefix, mmap_str );
	fprintf( dwip->fp, "Name=%s", sodbp->mmap_name );
    }else if (sodbp->sodb_type & SODB_T_IPC)
    {
	if (pref)
	    vre_Disp_Prefix( dwip->fp, sodbp,
			    dwip->base, dwip->prefix, ipc_str );
	fprintf( dwip->fp, "Key=%ld  TokenID=%s  ID=%d\n",
		(long)sodbp->ipc_key,
		sane_char(sodbp->ipc_tok_id),
		sodbp->ipc_id );
	if (pref)
	    vre_Disp_Prefix( dwip->fp, sodbp,
			    dwip->base, dwip->prefix, ipc_str );
	fprintf( dwip->fp, "Name=%s", sodbp->ipc_path );
    }else
    {
	fprintf( dwip->fp, "Unknown type!!" );
    }

    return(0);
}

static char *
build_time_t_str( time_t when, char *buf )
{
    struct timeval tmptime;
    static char tmpbuf[40];
    tmptime.tv_sec = when;
    tmptime.tv_usec = 0L;
    return(build_timestamp_str( &tmptime, (buf)?buf:tmpbuf ) );
}

static int
shm_proc_list_disp( struct shm_proc_list *spp,
		struct vre_disp_walk_item *dwip )
{
    int pref;
    const char *cp;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Me=%08lx  SerNbr=%d  Updated=%s\n",
	    spp->lno_me, spp->serial_nbr,
	    build_time_t_str( spp->update_tm, NULL ) );
    cp = stack_prefix( dwip, ".List" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &spp->proc_List,
			dwip->base, dwip->prefix, NULL );
    bpo_list_disp( &spp->proc_List, dwip );
    cp = unstack_prefix( dwip, cp );
    return(0);
}

static int
shm_process_disp( struct shm_process *spp,
		    struct vre_disp_walk_item *dwip )
{
    int pref;
    const char *cp;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Me=%08lx  Pid=%s  Started=%s\n",
	    spp->lno_me, spp->pid_str,
	    build_time_t_str( spp->pid_start_tm, NULL ) );
    if (pref)
	vre_Disp_Prefix( dwip->fp, spp,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Name=%s  Group=%s  Fifo=%s\n",
	    lno_string( spp->pid_nm_o, spp ),
	    lno_string( spp->pid_grp_o, spp ),
	    lno_string( spp->pid_fifo_o, spp ) );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &spp->proc_Node,
			dwip->base, dwip->prefix, ".proc_Node" );
    bpo_node_disp( &spp->proc_Node, dwip );
    cp = stack_prefix( dwip, ".pid_resources" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &spp->pid_resources,
			dwip->base, dwip->prefix, NULL );
    bpo_list_disp( &spp->pid_resources, dwip );
    cp = unstack_prefix( dwip, cp );
    return(0);
}

static int
q_head_disp( struct Q_Head *q_head,
	    struct vre_disp_walk_item *dwip )
{
    int pref;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Me=%08lx\n", q_head->lno_me );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &q_head->Qh_List,
			dwip->base, dwip->prefix, NULL );
    bpo_list_disp( &q_head->Qh_List, dwip );
    return(0);
}

static int
q_pid_disp( struct bpo_Q_pid *qpp,
	    struct vre_disp_walk_item *dwip )
{
    int pref;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Pid=%s\n", qpp->Q_pidstr );
    if (pref)
	vre_Disp_Prefix( dwip->fp, &qpp->Q_pid_Node,
			dwip->base, dwip->prefix, NULL );
    bpo_node_disp( &qpp->Q_pid_Node, dwip );
    return(0);
}

static int
q_disp( struct bpo_Q *qp,
	struct vre_disp_walk_item *dwip )
{
    int pref;
    const char *cp;
    pref = (dwip->flags & VRE_DISP_ELEM_NAMES);
    fprintf( dwip->fp, "Name=%s\n", qp->Q_name );
    if (pref)
	vre_Disp_Prefix( dwip->fp, qp,
			dwip->base, dwip->prefix, NULL );
    fprintf( dwip->fp, "Count=%d  Max=%d  Notify=%d  Flags=%d\n",
	    qp->Q_Count, qp->Q_max_count, qp->Q_notify, qp->Q_flags );
    cp = stack_prefix( dwip, ".Q_pid_r_list" );
    if (pref)
	vre_Disp_Prefix( dwip->fp, qp,
			dwip->base, dwip->prefix, NULL );
    bpo_list_disp( &qp->Q_pid_r_list, dwip );
    cp = unstack_prefix( dwip, cp );
    return(0);
}

int
sh_diag_disp_init( void )
{
    vre_New_Type( "bpo_Node_t",
		(vre_disp_t)bpo_node_disp,
		sizeof(bpo_Node_t),
		MPLIB_ALIGN_SPC_INT, 1 );
    vre_Type_Synonym( "bpo_Node_t", "bpo_Node" );
    vre_New_Type( "bpo_List_t",
		(vre_disp_t)bpo_list_disp,
		sizeof(bpo_List_t),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_Type_Synonym( "bpo_List_t", "bpo_List" );
    vre_New_Type( "b_lock_t",
		(vre_disp_t)b_lock_disp,
		sizeof(b_lock_t),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_Type_Synonym( "b_lock_t", "b_lock" );
    vre_New_Type( "bpo_pid_lock_t",
		(vre_disp_t)bpo_pid_lock_disp,
		sizeof(struct bpo_pid_lock),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_Type_Synonym( "bpo_pid_lock_t", "bpo_pid_lock" );

    vre_New_Type( "sodb_resource",
		(vre_disp_t)sodb_resource_disp,
		sizeof(struct sodb_resource),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "sodb",
		(vre_disp_t)sodb_disp,
		sizeof(struct sodb),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "shm_proc_list",
		(vre_disp_t)shm_proc_list_disp,
		sizeof(struct shm_proc_list),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "shm_process",
		(vre_disp_t)shm_process_disp,
		sizeof(struct shm_process),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "Q_Head",
		(vre_disp_t)q_head_disp,
		sizeof(struct Q_Head),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "bpo_Q_pid",
		(vre_disp_t)q_pid_disp,
		sizeof(struct bpo_Q_pid),
		MPLIB_ALIGN_SPC_LONG, 1 );
    vre_New_Type( "bpo_Q",
		(vre_disp_t)q_disp,
		sizeof(struct bpo_Q),
		MPLIB_ALIGN_SPC_LONG, 1 );
    return(0);
}


/* -- End of File -- */
